/*
 * Copyright 2001 Nimrod Zimerman. All Rights Reserved.
 * Release 20010630.
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * SECURITY NOTE:
 *
 * This program is a SECURITY HAZARD. It should have never been written.
 *
 * Every attempt has been made to make it as secure as possible, but
 * mistakes occur.
 * Read it. Understand it. If you find any problems, no matter how small,
 * let me know.
 *
 * Even if the code in this program is completely safe and has no
 * security problems whatsoever, it is still possible that operating system
 * bugs, library bugs, utility bugs, compiler bugs, processor bugs and
 * alike would cause it to work incorrectly.
 *
 * Be afraid.
 *
 *                                           Nimrod Zimerman,
 *                                           zimerman@mailandnews.com.
 */

/* CONSTANTS TO BE CHANGED: */
/* (Change the constants below with great care!) */

/* Path to the 'useradd' binary. */
#define USERADD_PATH "/usr/sbin/useradd"

/* Path to the 'chfn' binary. */
#define CHFN_PATH "/usr/bin/chfn"

/* Path to the 'chpasswd' binary. */
#define CHPASSWD_PATH "/usr/sbin/chpasswd"

/* The only user allowed to execute the program: */
#define ALLOWED_UID 420

/* The only user the program is allowed to run as: */
#define ALLOWED_EUID 0

/* The location of all home directories (with a trailing slash!): */
#define HOMEBASE "/home/"

/* The shell assigned to newly created users: */
#define SHELL "/dev/null"

/* The skeleton home directory (files from it would be copied to the
   home directory): */
#define SKELDIR "/dev/null"

/* END OF CONSTANTS TO BE CHANGED. */


#define ALPHA_CHARS   "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define DIGIT_CHARS   "0123456789"
#define SPECIAL_CHARS " '\".()-"


/*
 * Validate the user permissions by making sure only the allowed
 * user is running the program. 
 */
void validate_user_permissions()
{
	uid_t uid  = getuid();
	uid_t euid = geteuid();

	if (uid != ALLOWED_UID)
	{
		fprintf (stderr, "The program can only be executed by user id %d. Sorry.\n",
		         ALLOWED_UID);
		exit (1);
	}
	
	if (euid != ALLOWED_EUID)
	{
		fprintf (stderr, "The program can only run as user id %d. Sorry.\n",
		         ALLOWED_EUID);
		exit (1);
	}

	if (setreuid (ALLOWED_EUID, ALLOWED_EUID) == -1)
	{
		fprintf (stderr, "Error setting real and effective user uids. Sorry.\n");
		perror ("setreuid");
		exit (1);
	}
}

/*
 * Parse the command line, returning username, name and password.
 */
void parse_cmd_line (int argc, char *argv[],
                     char **username, 
                     char **name, 
                     char **password)
{
	if (argc != 4)
	{
		fprintf (stderr, "Incorrent number of arguments. Sorry.\n");
		fprintf (stderr, "Usage: %s username name password\n", argv[0]);
		exit (1);
	}
	
	*username = argv[1];
	*name     = argv[2];
	*password = argv[3];
}

/*
 * Return true if 'chr' is one of the characters composing 'set'.
 */
int char_in_set (char chr, char *set)
{
	char *p = set;
	while (*p != '\0')
	{
		if (*p == chr)
			return 1;
		p++;
	}
	return 0;
}

/*
 * Return true if 'str' starts with a character from 'first_char_set'
 * and only contains characters from 'allowed_chars_set'.
 * Subsequently, if 'first_char_set' doesn't contain '\0', the
 * string isn't allowed to be empty.
 */
int validate_string (char *str,
                     char *first_char_set, char *allowed_chars_set)
{
	char *p = str;

	if (!char_in_set (*p, first_char_set))
		return 0;
		
	while (*p != '\0')
	{
		if (!char_in_set (*p, allowed_chars_set))
			return 0;
		p++;
	}
	return 1;
}

/*
 * Emit an error message indicating the input is invalid, and exit.
 */
void abort_input_invalid (char *input)
{
	fprintf (stderr, "Input for %s is invalid. Sorry.\n", input);
	exit (1);
}

/*
 * Validate that the username, name and passwords inputs consist
 * of allowed characters only.
 */
void validate_input (char *username, char *name, char *password)
{
	if (!validate_string (username,
	                      ALPHA_CHARS,
	                      ALPHA_CHARS DIGIT_CHARS))
		abort_input_invalid ("username");
		
	if (!validate_string (name, 
	                      ALPHA_CHARS DIGIT_CHARS,
	                      ALPHA_CHARS DIGIT_CHARS SPECIAL_CHARS))
		abort_input_invalid ("name");
		
	if (!validate_string (password,
	                      ALPHA_CHARS DIGIT_CHARS,	                   
	                      ALPHA_CHARS DIGIT_CHARS))
		abort_input_invalid ("password");
}

/*
 * Fork, making sure everything is okay and exiting otherwise.
 */
pid_t fork_with_check()
{
	pid_t pid = fork();

	if (pid == -1)
	{
		fprintf (stderr, "Fork failed. Sorry.\n");
		exit (1);
	}
	
	return pid;
}

/*
 * Wait for a process with the given pid to terminate, making sure
 * everything is okay and exiting otherwise.
 */
void waitpid_with_check (pid_t pid)
{
	int ret_pid;
	int status;
	int exit_status;

	ret_pid = waitpid (pid, &status, 0);
	if (ret_pid == -1)
	{
		fprintf (stderr, "Problem waiting for process to terminate. Sorry.\n");
		perror ("waitpid");
		exit (1);
	}
	if (ret_pid != pid)
	{
		fprintf (stderr, "Incorrect process returned from wait. Sorry.\n");
		exit (1);
	}

	if (!WIFEXITED(status))
	{
		fprintf (stderr, "Child process exited abnormally. Sorry.\n");
		exit (1);
	}
		
	exit_status = WEXITSTATUS(status);

	if (exit_status != 0)
	{
		fprintf (stderr, "Child process exited with an error. Sorry.\n");
		exit (1);
	}
}

/*
 * Write 'count' bytes from 'buf' to 'fd', making sure everything
 * is okay and exiting otherwise.
 */
void write_with_check (int fd, char *buf, size_t count)
{
	size_t bytes_written;

	bytes_written = write (fd, buf, count);
	
	if (bytes_written == -1)
	{
		fprintf (stderr, "Error while writing to pipe. Sorry.\n");
		perror ("write");
		exit (1);
	}
	
	if (bytes_written != count)
	{
		fprintf (stderr, "Short write on pipe. Sorry.\n");
		exit (1);
	}
}

/*
 * Close the fd 'fd', making sure everything is okay and exiting otherwise.
 */
void close_with_check (int fd)
{
	if (close(fd) == -1)
	{
		fprintf (stderr, "Error closing pipe. Sorry.\n");
		perror ("close");
		exit (1);
	}
}

/*
 * Emit an error message indicating an exec() failed, and exit.
 */
void abort_exec_failed()
{
	fprintf (stderr, "Execution of program failed. Sorry.\n");
	perror ("execl");
	exit (1);
}

/*
 * Execute the 'useradd' binary, creating a user named 'username',
 * with a given home directory and shell. Exit on failure.
 */
void exec_useradd (char *username)
{
	pid_t pid = fork_with_check();
	if (pid == 0)
	{
		char *homedir = malloc (strlen(HOMEBASE) + strlen(username) + 1);
		if (homedir == NULL)
		{
			fprintf (stderr, "Not enough memory while creating homedir string. Sorry.\n");
			exit (1);
		}
	
		strcpy (homedir, HOMEBASE);
		strcat (homedir, username);
		
		execl (USERADD_PATH, USERADD_PATH,
		       "-d", homedir,
		       "-s", SHELL,
		       "-m",
		       "-k", SKELDIR,
		       username,
		       NULL);

		abort_exec_failed();
	}
	waitpid_with_check (pid);
}

/*
 * Execute the 'chfn' binary, changing the user name of the given
 * user to 'name'. Exit on failure.
 */
void exec_chfn (char *username, char *name)
{
	pid_t pid = fork_with_check();
	if (pid == 0)
	{
		execl (CHFN_PATH, CHFN_PATH,
		       "-f", name,
		       username,
		       NULL);
		       
		abort_exec_failed();
	}
	waitpid_with_check (pid);
}

/*
 * Execute the 'chpasswd' binary, changing the password of the given
 * user to 'password'. Exit on failure.
 */
void exec_chpasswd (char *username, char *password)
{
	int fds[2];
	pid_t pid;
	char colon[] = ":";
	char cr[] = "\n";
	
	if (pipe (fds) == -1)
	{
		fprintf (stderr, "Failed creating a pipe. Sorry.\n");
		exit (1);
	}

	pid = fork_with_check();
	if (pid == 0)
	{
		close_with_check (fds[1]);
		close_with_check (0);

		if (dup2 (fds[0], 0) == -1)
		{
			fprintf (stderr, "Error duplicating file descriptor for standard input. Sorry.\n");
			perror ("dup2");
			exit (1);
		}

		execl (CHPASSWD_PATH, CHPASSWD_PATH,
		       NULL);
		
		abort_exec_failed();
	}
	
	close_with_check (fds[0]);

	write_with_check (fds[1], username, strlen(username));
	write_with_check (fds[1], colon,    strlen(colon));
	write_with_check (fds[1], password, strlen(password));
	write_with_check (fds[1], cr,       strlen(cr));
	close_with_check (fds[1]);
	
	waitpid_with_check (pid);
}

/*
 * Create a user as indicated from the command line.
 */
int main (int argc, char *argv[])
{
	char *username;
	char *name;
	char *password;
	
	validate_user_permissions();
	
	parse_cmd_line (argc, argv, &username, &name, &password);
	validate_input (username, name, password);
	
	exec_useradd  (username);
	exec_chfn     (username, name);
	exec_chpasswd (username, password);

	return 0;
}
